class PublicKey:
    def __init__(self:object, p:object, curve:object) -> None:
        self.p = p
        self.curve = curve

    def to_string (self:object, encoded=False) -> str:
        base = 2 * (1 + len("%x" % self.curve.n))

        x = str(hex(self.p.x)).zfill(base)
        y = str(hex(self.p.y)).zfill(base)

        if encoded:
            return "0004" + x + y
        return x + y 

    def from_strin(self:object) -> str:
        pass

    def to_der (self:object) -> str:
        pass

    def from_der (self:object) -> str:
        pass

    def __str__(self:object) -> str:
        return f"curve: {self.curve.name}\npoint: {self.p}"



class Signature:
    def __init__(self:object, r:int, s:int, r_id:int):
        self.r = r
        self.s = s
        self.r_id = r_id

    def __str__(self:object) -> str:
        return f"r: {self.r} \ns: {self.s}\nr_id: {self.r_id}"



from ecdsa.point import Point, INF_POINT

import math

# Helper functions 
def reduce_mod(x:int, p:int) -> int:
    return x % p

def equal_mod(x:int, y:int, p:int) -> bool:
    return reduce_mod(x - y, p) == 0

def inverse_mod(x:int, p:int) -> int:
    if reduce_mod(x, p) == 0:
        return None
    return pow(x, p-2, p)

"""
Eliptic curve formula
y^2 = x^3 + ax + b
"""
class Curve:
    """
    params:
        name -> curve name
        a, b -> curve paramets
        p    -> field(modulo p)
        g    -> generator point
        n    -> order of G (number of points generate by scalar multiplication using the point g)
    """
    def __init__(self, name:str, a:int, b:int, p:int, g:Point, n:int) -> None:
        self.name = name
        self.a = a
        self.b = b
        self.p = p
        self.g = g
        self.n = n

    # https://es.wikipedia.org/wiki/Criptograf%C3%ADa_de_curva_el%C3%ADptica
    def add(self, p1:Point, p2:Point) -> Point:
        if p1 == INF_POINT:
            return p2
        if p2 == INF_POINT:
            return p1

        x1, y1 = p1.x, p1.y
        x2, y2 = p2.x, p2.y

        # Vertical points
        if equal_mod(x2, x1, self.p) and equal_mod(y2, -y1, self.p):
            return INF_POINT

        if equal_mod(x1, x2, self.p) and equal_mod(y1, y2, self.p):
            u = reduce_mod((3*(x1**2) + self.a), self.p) * inverse_mod((2 * y1), self.p)
        else:
            u = reduce_mod((y2 - y1), self.p) * inverse_mod((x2 - x1), self.p)

        v  = reduce_mod((y1 - u * x1), self.p)
        x3 = reduce_mod((u**2 - x1 - x2), self.p)
        y3 = reduce_mod((-u * x3 - v), self.p) 

        return Point(x3, y3)

    # https://www.youtube.com/watch?v=u1VRbo_fhC8 | double-and-Add
    def scalar_multiplication (self:object, k:int, p:Point) -> Point:
        q = INF_POINT
        while k != 0:
            if k & 1 != 0:
                q = self.add(q, p)
            p = self.add(p, p)
            k >>= 1
        return q
    
    def contain (self, p:Point) -> bool:
        return equal_mod(p.y**2, p.x**3 + self.a * p.x + self.b, self.p)

secp256k1 = Curve(
    "secp256k1", 
    0x0000000000000000000000000000000000000000000000000000000000000000, # A
    0x0000000000000000000000000000000000000000000000000000000000000007, # B
    0xfffffffffffffffffffffffffffffffffffffffffffffffffffffffefffffc2f, # P
    Point(                                                              # G
        0x79be667ef9dcbbac55a06295ce870b07029bfcdb2dce28d959f2815b16f81798, # G.x    
        0x483ada7726a3c4655da4fbfc0e1108a8fd17b448a68554199c47d08ffb10d4b8, # G.y
    ),
    0xfffffffffffffffffffffffffffffffebaaedce6af48a03bbfd25e8cd0364141  # N
)

p224 = Curve(
    "p224",
    -3,                                                                     # A
    18958286285566608000408668544493926415504680968679321075787234672564,   # B
    26959946667150639794667015087019630673557916260026308143510066298881,   # P
    Point(                                                                  # G
        19277929113566293071110308034699488026831934219452440156649784352033,   # G.x
        19926808758034470970197974370888749184205991990603949537637343198772,   # G.y
    ),
    26959946667150639794667015087019625940457807714424391721682722368061    # N
)



from ecdsa.point import INF_POINT
from ecdsa.public_key import PublicKey
from ecdsa.private_key import PrivateKey
from ecdsa.curves import reduce_mod, inverse_mod
from ecdsa.signature import Signature

from hashlib import sha256
from random import randint

# Error codes
ERROR_NOT_SAME_HASH_FUNCTION       = 0x0001
ERROR_NOT_SAME_CURVE               = 0x0002
ERROR_SIGNATURES_VALUES_NOT_VALID  = 0x0003
ERROR_PRIVATE_KEY_VALUES_NOT_VALID = 0x0004

# https://en.wikipedia.org/wiki/Elliptic_Curve_Digital_Signature_Algorithm
def sign (private_key:PrivateKey, message:bytes) -> Signature:
    curve  = private_key.curve
    secret = private_key.secret

    e = int(sha256(message).hexdigest(), 16)
    r, s = 0, 0
    while r == 0 or s == 0:
        k  = randint(1, curve.n-1)
        kp = curve.scalar_multiplication(k, curve.g)
        
        r  = reduce_mod(kp.x, curve.n)
        s  = reduce_mod(inverse_mod(k, curve.n) * (e + r*secret), curve.n)
    
    r_id = kp.y & 1
    if kp.y > curve.n:
        r_id += 2
    
    return Signature(r, s, r_id)

def verify (public_key:PublicKey, message:bytes, signature:Signature) -> bool:
    curve = public_key.curve
    q = public_key.p
    
    r = signature.r
    s = signature.s

    e = int(sha256(message).hexdigest(), 16)

    iv = inverse_mod(s, curve.n)
    u1 = reduce_mod(e * iv, curve.n)
    u2 = reduce_mod(r * iv, curve.n)

    p1 = curve.scalar_multiplication(u1, curve.g)
    p2 = curve.scalar_multiplication(u2, q)

    p3 = curve.add(p1, p2)
    
    return p3.x % curve.n == r



class File:
    @classmethod
    def read(path:str) -> str:
        with open(path, "r") as file:
            return file.read()

    @classmethod
    def write(path:str, message:str) -> None:
        with open(path, "w") as file:
            file.write(message)

    @classmethod
    def append(path:str, message:str) -> None:
        with open(path, "a") as file:
            file.write(message)

    @classmethod
    def get_size(path:str) -> int:
        pass



from ecdsa.curves import secp256k1
from ecdsa.public_key import PublicKey

from ecdsa.utils import File

from random import randint

class PrivateKey():
    def __init__(self, curve=secp256k1, secret=None):
        self.curve = curve
        self.secret = secret or randint(1, curve.n-1)

    def generate_public_key(self):
        public_key_point = self.curve.scalar_multiplication(self.secret, self.curve.g)
        public_key = PublicKey(public_key_point, self.curve)

        return public_key

    def load_from_pem_file(self, path:str):
        pem_file_data = File.read(path)
        pem = None # base64.decode(pem_file.split("\n")[0])

    def to_string(self) -> str:
        pass

    def to_pem(self) -> str:
        pass

    def __str__(self):
        return f"curve:  {self.curve.name}\nsecret: {hex(self.secret)}"



from setuptools import setup, find_packages

VERSION = "0.1.0"
DESCRIPTION = "Minimalistic ecdsa"

setup(
    name="light-ecdsa",
    version=VERSION,
    author="0xFrijolito",
    author_email="<dante_torobolino@protonmail.com>",
    description=DESCRIPTION,
    packages=find_packages(),
    install_requires=[],
    keywords=["python", "ecdsa", "signatures"],
    classifies=[
        "Development :: 1 - Planning",
        "Intented Audience :: Developers",
        "Programming Lenguage :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows"
    ]
)
